/*-------------------------------------------------
 *    Kernel version 2.4.20 - 2.6.8, 2.6.13
 *    RS485 (LPT-port)
 * 
 *    Andre Zaitsev, Olga Voziakova
 *    ovoz@sai.msu.ru
 *    $History:
 *
 *---------------------------------------- */
#include <linux/module.h>
#include <linux/types.h>
#include <linux/kdev_t.h>
#include <linux/fs.h>
#include <linux/device.h>

#include <linux/errno.h>
#include <linux/slab.h>
#include <asm/uaccess.h>   //  copy_from_user
#include <linux/ioport.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <asm/irq.h>
#include <asm/io.h>
#include <linux/parport.h>

#include <mcua.h> // To define KERN_2_6
#ifdef KERN_2_6
   #include <linux/cdev.h>
   #include <linux/pnp.h>
#endif
#ifndef KERN_2_6_18
   #include <linux/devfs_fs_kernel.h>
#endif

#include <lpt.h>
#include <rs485.h>
#include <rs485_tools.h>
#include <rs485_irq.h>
#include <rserrno.h>
#include <rs485_fifo.h> 

static unsigned int major;
RS485INFO lpt_table[PARPORT_MAX];
#ifndef __PP_PC__
static int __initdata io[PARPORT_MAX] = {_IO_LPT1, [1 ... PARPORT_MAX-1]=_IO_LPT2 };
static int __initdata irq[PARPORT_MAX] = {_IRQ_LPT1, [1 ... PARPORT_MAX-1]=_IRQ_LPT2};
#endif
static int __initdata baudrate[PARPORT_MAX] = {_BAUDRATE_1, [1 ... PARPORT_MAX-1]=_BAUDRATE_2};


struct file_operations rs485_fops;
static volatile int data_ready = 0;
DECLARE_WAIT_QUEUE_HEAD(tr_queue); // Test com-porta pokazal, chto eto medlennee prostogo oprosa (kak dlia 2.4)

#ifdef KERN_2_6
#ifndef __PP_PC__
// S PnP svyazyvaemsia tol'ko dlya togo, chtoby activirovat' parport, inache 2.6.18 ne rabotaet..
// Sodrano s parport_pc.c
static const struct pnp_device_id rs_lpt_pnp_tbl[] = {
        /* Standard LPT Printer Port */
        {.id = "PNP0400", .driver_data = 0},
        /* ECP Printer Port */
        {.id = "PNP0401", .driver_data = 0},
        { }
};

MODULE_DEVICE_TABLE(pnp, rs_lpt_pnp_tbl);

static int rs_lpt_pnp_probe(struct pnp_dev *dev, const struct pnp_device_id *id)
{
    return 0;
}

static void rs_lpt_pnp_remove(struct pnp_dev *dev)
{ }


static struct pnp_driver rs_lpt_pnp_driver = {
    .name           = "parport_rs485",
    .id_table       = rs_lpt_pnp_tbl,
    .probe          = rs_lpt_pnp_probe,
    .remove         = rs_lpt_pnp_remove,
};

static int pnp_registered_parport;

void rs485_activate_parport(void) {
    int err;
    err = pnp_register_driver(&rs_lpt_pnp_driver);
    if(!err) pnp_registered_parport = 1;
}

void rs485_stop_parport(void) {
    if(pnp_registered_parport)
        pnp_unregister_driver(&rs_lpt_pnp_driver);
}
#endif
#endif

inline void rs485_allow_interrupt_card(RS485INFO *pinfo) {
    u8 a = inb(pinfo->base +EPP_CONTROL);
#ifdef __M_BARRIER__
    rmb();
#endif
    outb( a | EPP_INT_ENABLE, pinfo->base +EPP_CONTROL );
#ifdef __M_BARRIER__
    wmb();
#endif
}

inline void rs485_stop_interrupt_card(RS485INFO *pinfo) {
    u8 a = inb(pinfo->base +EPP_CONTROL);
#ifdef __M_BARRIER__
    rmb();
#endif
    outb( a & ~EPP_INT_ENABLE, pinfo->base +EPP_CONTROL );
#ifdef __M_BARRIER__
    wmb();
#endif
}

inline void rs485_transmit_disable(RS485INFO *pinfo) {
    u8 a = inb(pinfo->base +EPP_CONTROL);
#ifdef __M_BARRIER__
    rmb();
#endif
    outb( a | ENABLE_BI_DIRECT,
                pinfo->base +EPP_CONTROL );
#ifdef __M_BARRIER__
    wmb();
#endif
}

inline void rs485_transmit_enable(RS485INFO *pinfo) {
    u8 a = inb(pinfo->base +EPP_CONTROL);
#ifdef __M_BARRIER__
    rmb();
#endif
    outb( a & ~ENABLE_BI_DIRECT,
                        pinfo->base +EPP_CONTROL );
#ifdef __M_BARRIER__
    wmb();
#endif
}

int start_transmitter(RS485INFO *info) {

#ifndef KERN_2_6
    int ready=0;
    int resp_tmout=0;
#endif
    long remain_tm;
    const long timeout = HZ * RESP_TIMEOUT / 1000L / 1000L;    // JIFFIES

#ifdef __DEBUG_SIGNAL__
    outb(TXENABLE, 0x3f8 +UART_MCR);    // to COM1
#endif

    data_ready = 0;    // not ready
    rs485_transmit_enable(info);

    info->rs->usCountByteOut = 0;

    while( 1 ) {

        info->rs->usCountByteOut ++;
        if( info->rs->usCountByteOut >= info->rs->outbuffer.size ) {
            break;
        }

        outb(info->rs->outbuffer.cache[info->rs->usCountByteOut], // 1-st, address byte
                info->base +EPP_DATA);
#ifdef __M_BARRIER__
        wmb();
#endif
    }
    // send address (end of packet)
    outb( info->rs->outbuffer.cache[0], info->base +EPP_ADDR );
#ifdef __M_BARRIER__
    wmb();
#endif
    info->rs->status = PASSIVE_DATA;

    PDEBUGG( "send: %d bytes, QUIT\n",info->rs->usCountByteOut );

#ifndef KERN_2_6
    rs485_transmit_disable(info);

// waiting answer from module (in irq-handler)

    rs485_set_starttime();
    resp_tmout = 0;
    while( 1 ) {
//        rs485_stop_interrupts(); 
        disable_irq(info.irq); // protect race condition with handler
        ready = data_ready;
        enable_irq(info.irq);
//        rs485_allow_interrupts();

        if(ready) break;
        if( rs485_timeout(RESP_TIMEOUT) ) {
            resp_tmout = 1;
            break;
        }
        udelay(RESP_WAIT);
    }
#ifdef __DEBUG_SIGNAL__
    outb(RXENABLE, 0x3f8 +UART_MCR);    // to COM1
#endif
    if(resp_tmout) {
        PDEBUG( "transmit: timeout!\n" );
        return -ERSNOCONVREPLY;
    } else {
        return(info->rs->inbuffer_pas.size);
    }
#else
    remain_tm = wait_event_interruptible_timeout(tr_queue, data_ready != 0, timeout);
    rs485_transmit_disable(info);
#ifdef __DEBUG_SIGNAL__
    outb(RXENABLE, 0x3f8 +UART_MCR);    // to COM1
#endif
    if(signal_pending(current)) {
        return -ERESTARTSYS;
    }
    if( remain_tm == 0 ) {
        PDEBUG( "transmit: timeout!\n" );
        return -ERSNOCONVREPLY;
    }
    return(info->rs->inbuffer_pas.size);    // signal or packet length
#endif
}


void receiv_data(RS485INFO * info) {

/*
 * -----------------------------------------------------------
 *        Receiver
 *        if info->rs->status == PASSIVE_DATA, then it is answer to data request
 *        passive-data format: [length][data...]  [signal]
 *        active-data format: [address][length][data...]
 * -----------------------------------------------------------
*/

    u8 iir = 0;
    int count_byte = 0;
    int ret = SUCCESS;
    struct _IN_BUFFER_ *current_inbuf;

#ifdef __DEBUG_TIMING__
    rs485_set_starttime();
#endif

#ifdef __DEBUG_SIGNAL__
    outb(TXENABLE, 0x3f8 +UART_MCR);    // to COM1
#endif

START:    
    iir = inb(info->base +EPP_STATUS);
#ifdef __M_BARRIER__
    rmb();
#endif
    if( (iir & PARPORT_STATUS_ACK) == 0 ) {
        goto QUIT_HANDLER;
    }

    info->rs->signal = 0;
    info->rs->usCountByteInp = 0;
    if( info->rs->status == PASSIVE_DATA ) {
        current_inbuf = &(info->rs->inbuffer_pas);
        current_inbuf->size = 0;    // initiate size
        info->rs->lSection = SECT_LENGTHBYTE;
    } else {
        current_inbuf = &(info->rs->inbuffer_act);
        info->rs->lSection = SECT_ADDRESSBYTE;
    }

    while( 1 ) {

        info->rs->tmp_byte = inb( info->base +EPP_DATA );

        count_byte++;
        if(count_byte > CACHE_MAXSIZE) {
                PDEBUG("long handler\n");
                goto QUIT_HANDLER;
        }
                                                                                
        // check time-out converter-bus
/*
        iir = inb(info->base +EPP_STATUS);
        mb();
        if( iir & EPP_TIMEOUT ) {

            PDEBUG( "receiver: timeout\n" );
            // reset:
            outb(EPP_TIMEOUT, info->base +EPP_STATUS);
            // wmb();
            if( (iir & PARPORT_STATUS_ACK) == 0 ) {
                PDEBUG( "receiver: timeout: No data in converter\n" );
                goto QUIT_HANDLER;    // no data in converter
            }
            info->rs->tmp_byte = inb( info->base +EPP_DATA );
            mb();
            iir = inb(info->base +EPP_STATUS);
            mb();
            if( iir & EPP_TIMEOUT ) {
                PDEBUG( "receiver: repeat. timeout, quit\n" );
                goto QUIT_HANDLER;
            }
        }
*/

        switch(info->rs->lSection) {

        case SECT_ADDRESSBYTE:

            // no address-byte in passive-bufer

            current_inbuf->cache[info->rs->usCountByteInp] = info->rs->tmp_byte;
            info->rs->usCountByteInp ++;
            current_inbuf->adr  = info->rs->tmp_byte;

            // next byte is length
            info->rs->lSection = SECT_LENGTHBYTE;

            break;

        case SECT_LENGTHBYTE:

            // length or signal?
            if(info->rs->tmp_byte >= PACKET_MAX_LENGTH) { // not signal
                if(info->rs->tmp_byte < CACHE_MAXSIZE) { // length error
                    PDEBUG( "receiver: length error! length = %x\n", info->rs->tmp_byte);
                } else {    // signal
                    if(info->rs->status == PASSIVE_DATA ) {
                        current_inbuf->size = info->rs->tmp_byte;
                        data_ready = 1;
                        wake_up(&tr_queue);
                        info->rs->status = ACTIVE_DATA;
                    } else {    // Active data:
                        PDEBUG( "active receiver: signal = %x from address = %x received!\n",
                                     info->rs->tmp_byte, current_inbuf->adr);
//  ne nuzhen on              info->rs->signal = (char)info->rs->tmp_byte;
//  vsio ravno pravilno       ret = rs485_put_data(info);
//  turbina ego ne prochtet   if(ret != SUCCESS) {
//                            PDEBUG("receiv_data: ret from rs485_put_data(SIGNAL) = %d\n", ret);
//                        }
                    }
                }
                goto START;
            }

            current_inbuf->cache[info->rs->usCountByteInp] = info->rs->tmp_byte;
            info->rs->usCountByteInp ++;

            if( info->rs->status == PASSIVE_DATA ) {
                current_inbuf->size = info->rs->tmp_byte+1;
            } else {
                // +address in this case
                current_inbuf->size = info->rs->tmp_byte+2;
            }

            info->rs->lSection = SECT_DATABYTES;

            break;

        case SECT_DATABYTES:            

            // receiving

            current_inbuf->cache[info->rs->usCountByteInp] = info->rs->tmp_byte;
            info->rs->usCountByteInp ++;
            if( info->rs->usCountByteInp < current_inbuf->size ) break;

            // packet is received

            if( info->rs->status == ACTIVE_DATA ) {
//                    printk("size=%d inp=%d\n",current_inbuf->size,info->rs->usCountByteInp);
                // put to FIFO
#ifdef __DEBUG_TIMING__
                if( rs485_timeout(100L) ) {    // 1 millisecond = 1000 usec
                    RS_PRINT( "Long packet: > 100 usec!\n" );
                }
#endif

                  rs485_put_data(info);
                  if(ret != SUCCESS) {
                      PDEBUG("receiv_data: ret from rs485_put_data(DATA) = %d\n", ret);
                  }

                goto START;    // test buffer
            } else {

                // for passive data:

                data_ready = 1;
                wake_up(&tr_queue);    // 2.4 ???
                info->rs->status = ACTIVE_DATA;
                goto START;
            }

        }
    } // infinite cycle

QUIT_HANDLER:

#ifdef __DEBUG_SIGNAL__
    outb(RXENABLE, 0x3f8 +UART_MCR); // to COM1
#endif

    return;
}

#ifdef __PP_PC__
#ifndef KERN_2_6_24
static void lpt_irq (int irq, void * private, struct pt_regs * unused) {
#else
static void lpt_irq (void * private) {
#endif
    RS485INFO* info = (RS485INFO *) private;
    receiv_data(info);
}
#endif

// -------------- FILE OPERARIONS --------------------------------------------


#define Fosc        14746
#define Minbaudrate 115
#define Maxbaudrate 1840

int ubrr_to_rate(int baudrate_code) {
    int a;
    a = baudrate_code + 1;
    if(a == 0) return 0;
    return( baudrate_code == 0x80 ? Fosc/8 : Fosc/16/a );
}

int rate_to_ubrr(int baudrate) {
    int baudrate_code;
    baudrate = baudrate < Minbaudrate ? Minbaudrate : baudrate;
    baudrate = baudrate > Maxbaudrate ? Maxbaudrate : baudrate;

    baudrate_code = baudrate > 1200 ? 0x80 : Fosc/16/baudrate-1;
//    baudrate_code = baudrate > 1200 ? 0x80 : iround(Fosc/16.0/baudrate)-1;
//VK new converter 1800-speed has code -1 ( 0xFF )
    return baudrate_code;
}

int rs485_set_baudrate(RS485INFO *pinfo) {
    int baudrate_code;
    baudrate_code = rate_to_ubrr(pinfo->baudrate);
    pinfo->baudrate = ubrr_to_rate(baudrate_code);
    PDEBUG("baudrate=%d\n",pinfo->baudrate);

    if( pinfo->rs->interface_type == CONV_VERS1 ) {
//         udelay(10000);   
        pinfo->rs->outbuffer.size = 9;
        pinfo->rs->outbuffer.cache[pinfo->rs->outbuffer.size] = 0;
        pinfo->rs->outbuffer.cache[0] = 0x1f;
        pinfo->rs->outbuffer.cache[1] = 0x07;
        pinfo->rs->outbuffer.cache[2] = 0x1f;
        pinfo->rs->outbuffer.cache[3] = 0x00;
        pinfo->rs->outbuffer.cache[4] = baudrate_code;
        pinfo->rs->outbuffer.cache[5] = 0x73;
        pinfo->rs->outbuffer.cache[6] = 0x00;
        pinfo->rs->outbuffer.cache[7] = 0x01;
        pinfo->rs->outbuffer.cache[8] = 0x6c;
        start_transmitter(pinfo);
    }
    if( pinfo->rs->interface_type == CONV_VERS2 ) {
        pinfo->rs->outbuffer.size = 3;
        pinfo->rs->outbuffer.cache[pinfo->rs->outbuffer.size] = 0;
        pinfo->rs->outbuffer.cache[0] = 0x1f;
        pinfo->rs->outbuffer.cache[1] = 0x43;
        pinfo->rs->outbuffer.cache[2] = baudrate_code;
          start_transmitter(pinfo);
    }
//    udelay(10000);
    return SUCCESS;
}

int rs485_test_device(RS485INFO *pinfo) {

    u8 iir = inb(pinfo->base+EPP_STATUS);
#ifdef __M_BARRIER__
    rmb();
#endif
    if( iir & PARPORT_STATUS_PAPEROUT ) {
        PDEBUG( "NO Converter!\n");
        return -ERSNOINTERFACE;
    }

    if( iir & PARPORT_STATUS_SELECT ) {
        PDEBUG( "NO POWER\n" );
        return -ERSSHUTDOWN;
    }

    if( iir & EPP_TIMEOUT ) {
        PDEBUG( "EPP TIMEOUT\n" );
        outb( EPP_TIMEOUT, pinfo->base +EPP_STATUS );
#ifdef __M_BARRIER__
        wmb();
#endif
//        return -ERSISATMOUT; chtoby ne sbivat' s tolku programmy
    }
    return SUCCESS;
}

int rs485_reset(RS485INFO *pinfo) {
    u8 i;

    int ret = SUCCESS;
    ret = rs485_test_device(pinfo); // a to set_baudrate povesip komp!
    if(ret != SUCCESS) return ret;

    outb( EPP_TIMEOUT, pinfo->base +EPP_STATUS );
#ifdef __M_BARRIER__
    wmb();
#endif
    outb( 0, pinfo->base +EPP_CONTROL );                //reset
#ifdef __M_BARRIER__
    wmb();
#endif
    udelay(EPP_WAIT);    // 250 mksec
    outb( EPP_RESET, pinfo->base +EPP_CONTROL );
#ifdef __M_BARRIER__
    wmb();
#endif
    udelay(EPP_WAIT);    // 250 mksec

    outb( 0x0e, pinfo->base +EPP_DATA );
#ifdef __M_BARRIER__
    wmb();
#endif
    outb( 0x1f, pinfo->base +EPP_ADDR );
#ifdef __M_BARRIER__
    wmb();
#endif
    udelay(RESP_WAIT);    // 100 mksec
// New code:
    i = inb(pinfo->base+EPP_DATA);
#ifdef __M_BARRIER__
    rmb();
#endif
      if( i == SignalACN ) {
          pinfo->rs->interface_type = CONV_VERS2;
          outb( 0x87, pinfo->base +EPP_DATA ); //For new
#ifdef __M_BARRIER__
        wmb();
#endif
        outb( 0x1f, pinfo->base +EPP_ADDR );
#ifdef __M_BARRIER__
        wmb();
#endif
        udelay(RESP_WAIT);    // 100 mksec
        inb(pinfo->base+EPP_DATA);
#ifdef __M_BARRIER__
        rmb();
#endif
        PDEBUG("Converter code vers.2\n");
    } else  if (i == SignalACY) {
        pinfo->rs->interface_type = CONV_VERS1;
        PDEBUG("Converter code vers.1\n");
    } else {
        PDEBUG( "reset: NO POWER\n" );
        return -ERSSHUTDOWN;    // na nekotorykh kompakh test_device ne srabatyvayet
                        
    }

    udelay(EPP_WAIT);

    pinfo->rs->status = ACTIVE_DATA;
    rs485_transmit_disable(pinfo);
// Tut visnet komp naskvoz', esli ne vkliuchen pribor ili ne vkliuchen pribor ili net konvertera
// (Ne liuboj komp., no mnogie)

    rs485_allow_interrupt_card(pinfo);

    rs485_set_baudrate(pinfo);
    return ret;
}


void rs485_blnumber_reset(RS485INFO *info) {
    int i;
    for( i=0; i<MAX_TASK_DEVICE; i++ ) {
        info->rs->device[i].number = 0;
        info->rs->device[i].round_count_prev = -1;
    }
    info->rs->inbuffer_pas.adr = 0;
    info->rs->inbuffer_act.adr = 0;
}

int rs485_open(struct inode *inode, struct file *pfile) {
    int err = 0;
    unsigned int minor = iminor(inode);
    if(minor >= PARPORT_MAX) return -ENXIO;
    if((lpt_table[minor].flags & EXIST_FL) == 0) return -ENXIO;
    
    if( lpt_table[minor].use_mcua >= MAX_USE_STREAM ) return -ERSBUSY; // More than maximal shared users
    PDEBUG("%s: open device %d, %d\n", ADAPTER_ID, imajor(inode), iminor(inode));

    if( lpt_table[minor].use_mcua == 0 ) {    // First opening
        lpt_table[minor].rs = (struct RS *)kmalloc(sizeof(struct RS), GFP_KERNEL);
        if(!lpt_table[minor].rs) {
            return -ENOMEM;
        }
        rs485_stop_interrupt_card(&lpt_table[minor]);
        rs485_blnumber_reset(&lpt_table[minor]);
        rs485_fifo_init(&lpt_table[minor]);
// test_device sidit v reset(), t.k. v com nado snachala delat'
// reset, a potom test_device, a v lpt naoborot.
        err = rs485_reset(&lpt_table[minor]); 
        if( err < 0 ) {
            return err;
        }
    }

    lpt_table[minor].use_mcua ++;
#ifndef KERN_2_6
    MOD_INC_USE_COUNT;
#endif
    return 0;
}

int rs485_release(struct inode *inode, struct file *pfile) {

    RS485INFO *info = &(lpt_table[iminor(pfile->f_dentry->d_inode)]);

//   if( down_interruptible(&info->sem ) ) {
//       up(&info->sem);
//       return -EINTR;
//   }

#ifndef KERN_2_6
     MOD_DEC_USE_COUNT;
#endif
     info->use_mcua --;

     if( info->use_mcua == 0 ) {
         rs485_stop_interrupt_card(info);
         if(lpt_table[iminor(inode)].rs) {
             kfree(lpt_table[iminor(inode)].rs);
             lpt_table[iminor(inode)].rs = NULL;
         }
     }
//    udelay(SEM_WAIT);
//    up(&info->sem);

    return 0;
}

ssize_t rs485_write( struct file *fp, const char __user *buf, size_t size, loff_t *offset ) {

/*  Data write to port ( This isn't standard operation write )
 *  return:  < 0    - error code
 *           > 0x80 - signal
 *           < 0x20 - length of the packet
 */    
    RS485INFO    *info = &(lpt_table[iminor(fp->f_dentry->d_inode)]);
    unsigned long rest = 0;
    int i, ret = 0;

    if( down_interruptible( &info->sem ) ) {
        up(&info->sem);
        return -EINTR;
    }
    if( (size == 0) || (size > PACKET_MAX_LENGTH ) ) {
        PDEBUG( "rs485_write: size=%zd is invalid!\n", size);
        ret =  -ERSFORMAT;
        goto OUT;
    }
    if( (ret=rs485_test_device(info)) < 0 ) goto OUT;
    rest = copy_from_user( info->rs->outbuffer.cache, buf, size );
    if(rest != 0) {
        PDEBUG("rs485_write: copy_from_user rest=%ld\n", rest);
        ret = -EFAULT;
        goto OUT;
    }
    info->rs->outbuffer.size = size;
    for(i=0; i<MAX_HOP; i++) {
        ret = start_transmitter(info);
        if( (ret != SignalEER) && (ret != SignalNAK) ) break;
        msleep(4);
    }
    if( (ret > 0) && ( ret <= PACKET_MAX_LENGTH ) ) {
        rest = copy_to_user( (char*)buf, info->rs->inbuffer_pas.cache, ret);
        if( rest != 0 ) {
            PDEBUG( "rs485_write: copy_to_user rest=%ld\n", rest );
            ret = -EFAULT;
            goto OUT;
        }       
    }
OUT:
    udelay(SEM_WAIT);
    up(&info->sem);
    return ret;
}

ssize_t rs485_read( struct file *fp, char __user *buf,
            size_t count, loff_t *offset ) {
/* 
 * Read data fifo
 */    
    char count0;
    RS485INFO    *info = &(lpt_table[iminor(fp->f_dentry->d_inode)]);
    int ret = 0;

    if( down_interruptible(&info->sem)) {        // 2.4 ???
        up(&info->sem);
        return -EINTR;
    }

    ret = rs485_data_fifo_read( info, buf, 1 );
    if( ret > 0 ) {
        get_user(count0, buf);    // buf in User space!!!
        if( count0 < (count-1) ) count = count0;
        else count--;
        ret = rs485_data_fifo_read( info, buf+1, count );
    }
//        udelay(SEM_WAIT);
    up(&info->sem);
    return ret;
}


#ifdef KERN_2_6_36
long rs485_ioctl(struct file *fp,
            unsigned int cmd, unsigned long arg) {
#else
int rs485_ioctl(struct inode *inode, struct file *fp,
            unsigned int cmd, unsigned long arg) {
#endif

    int err = SUCCESS;
    RS485INFO *info;
    unsigned long rest = 0;
    
    if(_IOC_TYPE(cmd) != RS_IOC_MAGIC) {        // 2.4 ???
        PDEBUG("ioctl_error!\n");
        return -ENOTTY;
    }
    if(_IOC_NR(cmd) > RS_IOC_MAXNR) {
        PDEBUG("ioctl_error!\n");
        return -ENOTTY;
    }
#ifdef KERN_2_6
/*
    if(_IOC_DIR(cmd) & _IOC_READ) {
        acc = access_ok(VERIFY_WRITE, (void __user *)arg, _IOC_SIZE(cmd));
    } else {
        acc = access_ok(VERIFY_READ, (void __user *)arg, _IOC_SIZE(cmd));
    }

    if(!acc) {
        PDEBUG("ioctl_error!\n");
        return -EFAULT;
    }
*/
#endif
    info = &(lpt_table[iminor(fp->f_dentry->d_inode)]);

    switch(cmd) {
        case RS_RESET:
            if( down_interruptible(&info->sem)) {
                up(&info->sem);
                return -EINTR;
            }
            err = rs485_reset(info);
            rs485_fifo_flush(info);
            udelay(SEM_WAIT);
            up(&info->sem);
            return err;

        case RS_FIFO_FLUSH:
            if( down_interruptible(&info->sem)) {
                up(&info->sem);
                return -EINTR;
            }
            rs485_fifo_flush(info);
            udelay(SEM_WAIT);
            up(&info->sem);
            return SUCCESS;

        case RS_SET_BAUDRATE:
            if( down_interruptible(&info->sem)) {
                up(&info->sem);
                return -EINTR;
            }
            err = rs485_test_device(info);
            if(err >= 0 ) {
                info->baudrate = (int)arg;
                err = rs485_set_baudrate(info);
            }
            udelay(SEM_WAIT);
            up(&info->sem);
            return err;

        case RS_GET_BAUDRATE:
            return info->baudrate;

        case RS_BLNUMBER_RESET:
            rs485_blnumber_reset(info);
            return SUCCESS;

        case RS_GET_VERS:
            rest = copy_to_user((char*)arg, RS485_UTS_RELEASE, sizeof(RS485_UTS_RELEASE));
            if(rest != 0) {
                PDEBUG("rs485_ioctl: get_vers: copy_to_user rest=%ld\n", rest);
                return -EFAULT;
            }
            return RS485_VERSION_CODE;

        case RS_GET_DATAFIFO_SIZE:
            return SIZE_STREAM_FIFO;

        default: break;
    }
    return -ERSNOIOCTL;
}

struct file_operations rs485_fops = {
    .owner        = THIS_MODULE,
    .open         = rs485_open,
    .release      = rs485_release,
    .read         = rs485_read,
    .write        = rs485_write,
#ifdef KERN_2_6_36
    .unlocked_ioctl = rs485_ioctl,
#else
    .ioctl        = rs485_ioctl,
#endif
};


// ------------------------------- INIT / EXIT ----------------------------

#ifdef __PP_PC__
#ifdef KERN_2_6_18
static struct class *lptdev_class;
#else
static struct class_simple *lptdev_class;
#endif

static unsigned int lpt_count = 0;
static void lpt_attach(struct parport *port) {
    struct pardevice * pdev = NULL;
    PDEBUG("rs485lpt: port %s attached\n", port->name);
//    pdev = parport_register_device (port, ADAPTER_ID, NULL,
//                                NULL, lpt_irq, PARPORT_DEV_EXCL, &lpt_table[lpt_count]);
    pdev = parport_register_device (port, ADAPTER_ID, NULL,
                                NULL, lpt_irq, 0, &lpt_table[lpt_count]);
    if (!pdev) {
        printk (KERN_WARNING "%s: failed to register device!\n", ADAPTER_ID);
        return;// -ENXIO;
    }
//    parport_put_port (port);
    if(parport_claim(pdev)) {
        PDEBUG("parport is not claimed\n");
        parport_unregister_device(pdev);
        return; //-EBUSY
    }
    lpt_table[lpt_count].pdev = pdev;
    lpt_table[lpt_count].flags |= EXIST_FL;
    lpt_table[lpt_count].base = pdev->port->base;
    lpt_table[lpt_count].irq = pdev->port->irq;
    printk (KERN_DEBUG "%s: registered pardevice %d\n", ADAPTER_ID, lpt_count);

#ifdef KERN_2_6_27
    device_create(lptdev_class, port->dev, MKDEV(major, lpt_count), NULL,
                    "rs485lpt%d", lpt_count);
#else
 #ifdef KERN_2_6_24
    device_create(lptdev_class, port->dev, MKDEV(major, lpt_count),
                    "rs485lpt%d", lpt_count);
 #else
  #ifdef KERN_2_6_18
    class_device_create(lptdev_class, NULL, MKDEV(major, lpt_count), NULL,
                    "rs485lpt%d", lpt_count);
  #else
    class_simple_device_add(lptdev_class, MKDEV(major, lpt_count), NULL,
                    "rs485lpt%d", lpt_count);
    devfs_mk_cdev( MKDEV(major, lpt_count), S_IFCHR | S_IRUGO | S_IWUGO,
                     "mcua%d", lpt_count);
  #endif
 #endif
#endif
    printk(KERN_INFO "rs485lpt%d: using %s (%s).\n", port->number, port->name,
        (port->irq == PARPORT_IRQ_NONE)?"polling":"interrupt-driven");
    PDEBUG("rs485lpt attach io=0x%lX irq=%d\n", port->base, port->irq);
    lpt_count++;
}

static void lpt_detach(struct parport *port) {
    PDEBUG("rs485lpt: %s  detached\n", port->name);
}

static struct parport_driver lpt_driver = {
    .name           = ADAPTER_ID,
    .attach         = lpt_attach,
    .detach         = lpt_detach,
};

#endif

#ifndef __PP_PC__
static void rs485_release_resource(void) {
    int i;
    for(i=0; i<PARPORT_MAX; i++) {
        if(lpt_table[i].flags & EXIST_FL) {
            release_region(lpt_table[i].base, ADDRESS_REGION_SIZE);
            rs485_clean_irq(&lpt_table[i]);
//            PDEBUG("irq=%d cleaned\n", lpt_table[i].irq);
//            PDEBUG("region %x released\n", lpt_table[i].base);
            lpt_table[i].flags &= ~EXIST_FL;
        }
    }
    return;
}

void __init rs485_request_resource(void) { // IO ports & irqs initialization
    int i;
    for(i=0; i<PARPORT_MAX; i++) {
#ifndef KERN_2_6
        if(check_region(lpt_table[i].base, ADDRESS_REGION_SIZE) != 0 ) {
//            RS_PRINT( KERN_CRIT "%s: I/O port 0x%08lX in use.\n",
//                        ADAPTER_ID, (unsigned long) lpt_table[i].base);
            continue;
        }
        request_region(lpt_table[i].base, ADDRESS_REGION_SIZE, ADAPTER_ID);
#else
        if(!request_region(lpt_table[i].base, ADDRESS_REGION_SIZE, ADAPTER_ID)) {
//            RS_PRINT( KERN_CRIT "%s: I/O port 0x%08lX in use.\n",
//                            ADAPTER_ID, (unsigned long) lpt_table[i].base);
            continue;
        }
#endif
        PDEBUG("port=%d, base=0x%08lX requested\n", i, (unsigned long) lpt_table[i].base);
        if(rs485_request_irq(&lpt_table[i])) {
            release_region(lpt_table[i].base, ADDRESS_REGION_SIZE);
            printk("region 0x%08lX released\n", (unsigned long)lpt_table[i].base);
            continue;
        }
        PDEBUG("port=%d, irq=%d requested\n", i, lpt_table[i].irq);
        lpt_table[i].flags |= EXIST_FL;
    }
}
#endif

MODULE_AUTHOR("Andre Zaitsev, Olga Voziakova (ovoz@sai.msu.ru)");
MODULE_DESCRIPTION("rs485 lpt module");
#ifdef MODULE_LICENSE
    MODULE_LICENSE("GPL");
#endif

#ifndef KERN_2_6_18
static unsigned int tmpvar=0;
#else
#define tmpvar NULL
#endif
#ifndef __PP_PC__
MODULE_PARM_DESC(io, "Base I/O address (hex), default io1=_IO_LPT1, io2=_IO_LPT2");
module_param_array(io, int, tmpvar, 0);
MODULE_PARM_DESC(irq, "IRQ line,default irq1=_IRQ_LPT1, irq2=_IRQ_LPT2");
module_param_array(irq, int, tmpvar, 0);
#endif
MODULE_PARM_DESC(baudrate, "BAUD RATE UART port, default baudrate1=_BAUDRATE_1, baudrate2=_BAUDRATE_2");
module_param_array(baudrate, int, tmpvar, 0);

#ifndef KERN_2_6
EXPORT_NO_SYMBOLS;
#endif

int __init init_module(void) {
    int err = 0;
    unsigned int i;
    char version[]=RS485_UTS_RELEASE;

    for(i=0; i<PARPORT_MAX; i++) {
        lpt_table[i].rs = NULL;
        lpt_table[i].use_mcua = 0;
        lpt_table[i].flags = 0;
        init_MUTEX(&lpt_table[i].sem);
        lpt_table[i].baudrate = baudrate[i];
#ifndef __PP_PC__
        lpt_table[i].base = io[i];
        lpt_table[i].irq = irq[i];
#else
        lpt_table[i].base = 0;
        lpt_table[i].irq = -1;
        lpt_table[i].pdev = NULL;
#endif
//        PDEBUG("init port %d: baudrate=%d base=%x irq=%d\n",
//            i+1, lpt_table[i].baudrate, lpt_table[i].base, lpt_table[i].irq);
    }

#ifndef __PP_PC__
    rs485_request_resource();
#endif

    if ( (major=register_chrdev (0, ADAPTER_ID, &rs485_fops)) <= 0) {
        printk ("rs485lpt: unable to get major\n");
        err = -EIO;
        goto out_chrdev;
    }
#ifndef __PP_PC__
    rs485_activate_parport();
#else

#ifdef KERN_2_6_18
    lptdev_class = class_create(THIS_MODULE, ADAPTER_ID);
#else
    devfs_mk_dir("mcua");
    lptdev_class = class_simple_create(THIS_MODULE, ADAPTER_ID);
    if (IS_ERR(lptdev_class)) {
        err = PTR_ERR(lptdev_class);
        goto out_class ;
    }
#endif
    if (parport_register_driver(&lpt_driver)) {
        printk ("rs485lpt: unable to register with parport\n");
        err = -EIO;
        goto out_register_driver;
    }
#endif  // __PP_PC__
    RS_PRINT(KERN_INFO "%s module, version %s\n",
                ADAPTER_ID, version);
    for(i=0; i<PARPORT_MAX; i++) {
        if(lpt_table[i].flags & EXIST_FL) {
            RS_PRINT(KERN_INFO "%s: install card on the base 0x%X and irq=%d\n",
            ADAPTER_ID, lpt_table[i].base, lpt_table[i].irq);
        }
    }
    return 0;

//    rs485_fifo_destroy();
//out_fifo:
#ifdef __PP_PC__
    parport_unregister_driver(&lpt_driver);
out_register_driver:
 #ifdef KERN_2_6_18
    class_destroy(lptdev_class);
//    PDEBUG("class destroyed\n");
 #else
    class_simple_destroy(lptdev_class);
//    PDEBUG("class simple destroyed\n");
out_class:
    devfs_remove("mcua");
 #endif
#else
    rs485_stop_parport();
#endif
    unregister_chrdev(major, ADAPTER_ID);
out_chrdev:
#ifndef __PP_PC__
    rs485_release_resource();
#endif
    return err;
}

void __exit cleanup_module(void){
#ifndef __PP_PC__
    rs485_stop_parport();
    rs485_release_resource();
#else
    int i;
    parport_unregister_driver(&lpt_driver);

    for(i=0; i<PARPORT_MAX; i++) {
        if(!(lpt_table[i].flags & EXIST_FL)) continue;
        parport_release(lpt_table[i].pdev);
        parport_unregister_device(lpt_table[i].pdev);
        printk ("%s%x: unregistered pardevice %d\n",ADAPTER_ID, i, i);
#ifdef KERN_2_6_24
        device_destroy(lptdev_class, MKDEV(major, i));
#else
#ifdef KERN_2_6_18
        class_device_destroy(lptdev_class, MKDEV(major, i));
#else
        devfs_remove("mcua/%d",i);
        class_simple_device_remove(MKDEV(major, i));
#endif
#endif
//        PDEBUG("class device minor=%d destroyed\n",i);
        lpt_table[i].flags &= ~EXIST_FL;
    }
#ifdef KERN_2_6_18
    class_destroy(lptdev_class);
#else
    devfs_remove("mcua");
    class_simple_destroy(lptdev_class);
#endif
//    PDEBUG("class destroyed\n");
#endif // PP_PC
    unregister_chrdev (major, ADAPTER_ID);
    return;
}
