/************************************************************************
 *	  MOXA CP132
 *	
*************************************************************************/
#include <linux/errno.h>
#include <linux/major.h>
#include <linux/delay.h>
#include <asm/io.h>

#include <mcua.h>
#include <moxa132.h>
#include <rserrno.h>

#include <linux/ioport.h>

#include <rs485.h>

#include "rs485_tools.h"
#include "rs485_fifo.h"
#include "rs485_irq.h"

//extern	int io1, io2;
//extern	int irq1, irq2;
extern RS485INFO *info_port1, *info_port2;
//extern	int portnumb;
//extern 	struct file_operations rs485_fops;

extern	int _find_PCImoxa(void);


static unsigned short iir_stat=0; // pri sleduyuchem schityvanii izmenitsia PE-bit
				// (po opisaniyu moxy) => nado zapomnit' dlia receiver-a
inline int rs485_get_port(int irq) {
	iir_stat = inb(info_port1->base +UART_LSR);
	if( (iir_stat & UART_LSR_DR) != 0 ) {
		return 1;	// info_port1
	}
	iir_stat = inb(info_port2->base +UART_LSR);
	if( (iir_stat & UART_LSR_DR) != 0 ) {
		return 2;	// info_port2
	}
	return 0;	// lozhnoe preryvanie;
}

inline void rs485_allow_interrupt_card(RS485INFO *pinfo) {
//    
	outb(UART_IER_RDI, pinfo->base +UART_IER);
}

inline void rs485_stop_interrupt_card(RS485INFO *pinfo) {
//    
	outb(0, pinfo->base +UART_IER);
}

inline int rs485_test_device(RS485INFO *pinfo) {
	return 0;
}

inline void rs485_read_enable(RS485INFO *pinfo) {
	outb(UART_MCR_OUT2, pinfo->base +UART_MCR);	//   
}

inline void rs485_transmit_enable(RS485INFO *pinfo) {
	outb(UART_MCR_RTS, pinfo->base +UART_MCR);
}

inline void rs485_clean_reg(RS485INFO *pinfo) {
	unsigned short i;
	for(i=0; i<8; i++) outb(0x0, pinfo->base+i);
}

#define Fosc		14746
#define Minbaudrate	115
#define Maxbaudrate 921

int ubrr_to_rate(int baudrate_code) {
	if(baudrate_code == 0) return 0;
	return(Fosc/16/baudrate_code);
}

int rate_to_ubrr(int baudrate) {
	int baudrate_code;
	baudrate = baudrate < Minbaudrate ? Minbaudrate : baudrate;
	baudrate = baudrate > Maxbaudrate ? Maxbaudrate : baudrate;
//	baudrate_code = iround(Fosc/16.0/baudrate);
	baudrate_code = Fosc/16/baudrate;
	return baudrate_code;
}

int rs485_set_baudrate(RS485INFO *pinfo) {
	int baudrate_code = rate_to_ubrr(pinfo->baudrate);
	pinfo->baudrate = ubrr_to_rate(baudrate_code);
	outb( 0x80, pinfo->base +UART_LCR );	//  
	outw(baudrate_code, pinfo->base +UART_DLL); 
	outb(OTHER, pinfo->base +UART_LCR);	// .  
	return(SUCCESS);
}

int rs485_reset(RS485INFO *pinfo) {	//  -

	rs485_clean_reg(pinfo);

	rs485_read_enable(pinfo);
	rs485_allow_interrupt_card(pinfo);

	rs485_set_baudrate(pinfo);
	
	outb(UART_FCR_ENABLE_FIFO
					| UART_FCR_CLEAR_RCVR
					| UART_FCR_CLEAR_XMIT
					| UART_FCR_TRIGGER_8	//   fifo,
											//     - 
											//    1- 
					,pinfo->base +UART_FCR);  // fifo

//   
//	inb( pinfo->base +UART_LSR );// eto schityvanie nichego krome PE-bita ne meniauet
	inb( pinfo->base +UART_RX );

        return rs485_test_device(pinfo);
//	rs485_fifo_flush(pinfo);
}

// -----------------------------------------------------------------------

int start_transmitter(RS485INFO *info) {
/*   
 *   < 0
 *   > CACHE_MAXSIZE
 *      < CACHE_MAXSIZE
*/
	unsigned  short iir=0;
	int ret=99;

	int rbyte_tmout=0;

	static int SharedMaskInterrupt;
	SharedMaskInterrupt = UART_LSR_TEMT;

	rs485_stop_interrupts();
//	rs485_stop_interrupt_card(info);
//	up_flags(FTRANSMITTER, &info->lFlags);	//    

	outb(FIRST, info->base +UART_LCR);	// 
	rs485_transmit_enable(info);
	outb(info->outbuffer.cache[0], info->base); //   

	info->usHopOut = 0;	//  hop
	info->usHopInp = 0;	//  hop
	info->signal = 0;

	//   
	info->usCountByteOut = 0;	//    0
	info->usCountByteInp = 0;	//    0
	info->lSection = SECT_ADDRESSBYTE;	//   

	while( 1 ) {

		//  
		rs485_set_starttime();
		rbyte_tmout = 0;
		while( 1 ) {
			if( ( (iir=inb(info->base +UART_LSR)) & SharedMaskInterrupt ) != 0 ) break;
			if( rs485_timeout(TBYTE_TIMEOUT) ) {
				rbyte_tmout = 1;
				break;
			}
			udelay(TBYTE_WAIT);
		}


//       
		if( rbyte_tmout ) {

			switch(info->lSection) {

// no reply:
				case SECT_REPLYBYTE	:
					ret = SignalDSC;
									// no reply from module => disconnected
					PDEBUG( "transmit: timeout: no reply!\n" );
					goto QUIT_TRANSMITTER;

				case SECT_BADATA		:	//   bad-bytes,
											//  
					info->lSection = SECT_DATABYTES;
// data loss:
				case SECT_DATABYTES	:	//     
					(info->usHopInp) ++;	//   hop
					if(info->usHopInp < MAX_HOP) {
						info->signal = SignalNAK;
						PDEBUG( "transmit: broken packet!\n" );
						goto GOTO_SEND_SIGNAL;
					} else { 
						//  HOP   ACK   
						ret = SignalEER;
//						ret = -ERSTRANSMIT;	//    
						PDEBUG( "transmit: broken packet, MAX_HOP reached!\n" );
						goto QUIT_TRANSMITTER;
					}

				default	:	break;
			}
		}	// end of timeout-condition

		info->tmp_byte = inb(info->base);     //  

		if( (iir & UART_LSR_FE) != 0 ) {	//  .  , 
			PDEBUG( "transmit: frame error!\n" );

//			up_flags(FCOLLISION, &info->lFlags);
			switch(info->lSection) {

				case SECT_ADDRESSBYTE	:
				case SECT_SIGNAL		: 
					outb(OTHER, info->base +UART_LCR);	// . 
				case SECT_OUTBYTES	:
					rs485_read_enable(info);
				default: break;
			}
			ret = -ERSFRAME;
			goto QUIT_TRANSMITTER;
		}

		switch( info->lSection ) { //    
		
			case SECT_ADDRESSBYTE:

				//   
				//        OUTBYTES
				outb(OTHER, info->base +UART_LCR);
				info->lSection = SECT_OUTBYTES;

			case SECT_OUTBYTES:

				if( info->usCountByteOut == info->outbuffer.size ) {
					//   
					//    REPLYBYTE
					//   

					rs485_read_enable(info);
					SharedMaskInterrupt = UART_LSR_DR;

					info->lSection = SECT_REPLYBYTE;

				} else {

					//     
					(info->usCountByteOut) ++;
					outb(*(info->outbuffer.cache +info->usCountByteOut),
					 		info->base);
				}
				break;

			case SECT_REPLYBYTE: 

				//  

				if( iir  & UART_LSR_PE ) {	//  
					info->signal = info->tmp_byte;

// !!! Iz-za pomekh na linii vmesto ACY ingda prikhisit ACK, t.k. oni
// !!! otlochayutsia tol'ko odnim bitom.
// !!! Tut ACK ne mozhet byt', znachit, eto iskazhennyi ACY
//					if( info->signal == SignalACK ) info->signal = SignalACY;

					if( info->tmp_byte != (unsigned char) SignalNAK ) { 

						//     nak  
						//     
						ret = info->signal;
						goto QUIT_TRANSMITTER;

					} else {	//   nak

						PDEBUG( "transmit: NAK from module!\n" );
						info->usHopOut ++;
						if( info->usHopOut < MAX_HOP ){

							//       
							//   

							outb(FIRST, info->base +UART_LCR);	//  1- 
							rs485_transmit_enable(info);
							info->usCountByteOut = 0;
							SharedMaskInterrupt = UART_LSR_TEMT;
							info->lSection = SECT_ADDRESSBYTE;

							//  
							outb(info->outbuffer.cache[0], info->base);
							break;

						} else {

							//     

							ret = SignalNAK;
							goto QUIT_TRANSMITTER;
						}
					}

				} else { //    

					//  1   
					if( info->tmp_byte < (unsigned char)PACKET_MAX_LENGTH ) {

						info->passive_inbuf.size = info->tmp_byte +2;
						info->passive_inbuf.cache[info->usCountByteInp]=info->tmp_byte;
						info->lSection = SECT_DATABYTES;
						(info->usCountByteInp) ++;

					} else {
						PDEBUG( "transmit: bad length-byte, skipped!\n" );
						//     ,
						//    (NAK)
						info->lSection = SECT_BADATA;
					}
				}
				break;

			case SECT_BADATA:
				break;

			case SECT_DATABYTES: 
			
				//  

				info->passive_inbuf.cache[info->usCountByteInp] =
							info->tmp_byte; //    
				(info->usCountByteInp) ++;	

				if( info->usCountByteInp == info->passive_inbuf.size ) {

					//    -   
					if( crc(info->passive_inbuf.cache, 
							info->passive_inbuf.size) != 0 ) {

						//   
						PDEBUG( "transmit: bad CRC!\n" );
						(info->usHopInp) ++;	//   hop
						if(info->usHopInp == MAX_HOP) { 

							//  HOP  ACK   

							info->signal = SignalACK;
							ret = SignalEER;
//							ret = -ERSTRANSMIT;
													//  ?
						} else info->signal = SignalNAK; //   NAK
					} else { 
						//   
						//  ACK       

						info->signal = SignalACK;
						ret = info->passive_inbuf.size-1;	// size of data without CRC
					}
				}
				break;

			case  SECT_SIGNAL:

				//    
				//    

				outb(OTHER, info->base +UART_LCR);
				rs485_read_enable(info);

				if(info->signal == SignalNAK) {
					//     NAK,     
					info->signal = 0;
					info->lSection = SECT_REPLYBYTE;
					SharedMaskInterrupt = UART_LSR_DR;
					info->usCountByteInp = 0;
				} else {

					//  
					goto QUIT_TRANSMITTER;
				}
				break;

			default: break;

		}	//    

GOTO_SEND_SIGNAL:

		if( info->signal && (info->lSection == SECT_DATABYTES) ) {

			//    

			outb(FIRST, info->base +UART_LCR); //   
			rs485_transmit_enable(info);
			info->lSection = SECT_SIGNAL;
			SharedMaskInterrupt = UART_LSR_TEMT;
			outb(info->signal, info->base); 
		}

	}	//    

QUIT_TRANSMITTER:

//	down_flags(FTRANSMITTER, &info->lFlags);
//	rs485_allow_interrupt_card(info);
	rs485_allow_interrupts();

	return ret;
}

inline void receiv_data(RS485INFO * info) {

/*
 * -----------------------------------------------------------
 *		  
 * ------------------------------------------------------------
*/

	int rbyte_tmout=0;
 	unsigned short iir=0;
	int first = 1;

//#ifdef __DEBUG_TIMING__
//	static rs485_time_t hand_start_prev;
//	rs485_time_t		hand_start, hand_fin;
//#endif

	unsigned char current_address = 0;
	static int SharedMaskInterrupt = UART_LSR_DR;

#ifdef __DEBUG_TIMING__
	rs485_set_starttime();
#endif 

//	up_flags(FRECIVER, &info->lFlags);


//	rs485_stop_interrupt_card(info);


#ifdef __DEBUG_SIGNAL__
	outb(TXENABLE, info->base +RX_TX_REG + 8);
#endif

//#ifdef __DEBUG_TIMING__
//	rs485_gettime(&hand_start);
//	rs485_printtm_interval( &hand_start_prev, &hand_start );
//	hand_start_prev = hand_start;
//#endif

//	info->usHopInp = 0;
	info->signal = 0;
	info->lSection = SECT_ADDRESSBYTE;

	while( 1 ) {

//    
		rs485_set_starttime();
		rbyte_tmout = 0;
		if(first != 1) {	// gotovnost' uzhe schitana pri vhode v handler;
			while( 1 ) {
				iir = inb(info->base +UART_LSR);
				if( (iir & SharedMaskInterrupt) != 0 ) break;

				if( rs485_timeout(RBYTE_TIMEOUT) ) {
					rbyte_tmout = 1;
					break;
				}
				udelay(RBYTE_WAIT);
			}
		} else {
			first = 0;
			iir = iir_stat;
		}

//   - - 
		if( rbyte_tmout ) {
			PDEBUG( "receiver: time-out!\n" );
			if( info->lSection != SECT_SIGNAL ) goto QUIT_HANDLER;
		}

//  
		info->tmp_byte = inb(info->base);
//			PDEBUGG( "%X",info->tmp_byte );
			
//   ? 
		if( (iir & UART_LSR_FE) != 0 ) {
			PDEBUG( "receiver: frame error!\n" );
			goto QUIT_HANDLER;
		}

//   FIFO?
		if( (iir & UART_LSR_OE) != 0 ) {

			info->active_inbuf.signature |= S_FIFO_OVERFLOW;

			PDEBUG( "receiver: hard-FIFO overflow!\n" );

			//  FIFO
			outb( UART_FCR_ENABLE_FIFO | UART_FCR_CLEAR_RCVR
				| UART_FCR_CLEAR_XMIT
				| UART_FCR_TRIGGER_8	//   fifo,
										//     - 
										//    1- 
				,info->base +UART_FCR );

			goto QUIT_HANDLER;
		}

		switch( info->lSection ) {

		case SECT_ADDRESSBYTE:

			//        
			//   ,    (  :-))

			if( !(iir & UART_LSR_PE) ) {
				PDEBUG( "receiver: non address in address-section!\n" );
				goto QUIT_HANDLER;
			}

			//    - ,     
			if((unsigned char)(info->tmp_byte) >= (unsigned char) CACHE_MAXSIZE ){
				PDEBUG( "receiver: signal %x in address-section!\n",info->tmp_byte );
				goto QUIT_HANDLER;
			}

			//    
			if(info->active_inbuf.adr == info->tmp_byte) {

				PDEBUG( "receiver: repeated packet!\n" );
				//  ,    ,    ,
				//      
				// -      ...
				goto QUIT_HANDLER;
			}

			//   ,     
			info->usCountByteInp = 0;
			info->active_inbuf.cache[info->usCountByteInp]
					= info->tmp_byte;

			//   
			current_address = info->tmp_byte;
		                                                            
			//      
			info->usCountByteInp = 1;
			info->lSection = SECT_LENGTHBYTE;

			break;

		case SECT_LENGTHBYTE:

			//     or  - 
			if( iir & UART_LSR_PE ) {
				PDEBUG( "receiver: address or signal in length-section!\n" );
				goto QUIT_HANDLER;
			}

			//   ?
			if(info->tmp_byte >= 0x20) {
				PDEBUG( "receiver: length error!\n" );
				goto QUIT_HANDLER; //    
			}

			info->active_inbuf.cache[info->usCountByteInp]
					= info->tmp_byte;
			info->usCountByteInp ++;
			info->active_inbuf.size = info->tmp_byte+3;
			info->lSection = SECT_DATABYTES;

			break;

		case SECT_DATABYTES:			
			//  

			//     or  - 
			if( iir & UART_LSR_PE ) {
				PDEBUG( "receiver: address or signal in data-section!\n" );
				goto QUIT_HANDLER;
			}

			info->active_inbuf.cache[info->usCountByteInp]
					= info->tmp_byte;
			info->usCountByteInp ++;

			//       -   
			if( info->usCountByteInp < info->active_inbuf.size ) break;

			//    -     
			if( (crc(info->active_inbuf.cache,
				info->active_inbuf.size)) != 0 ) {

				PDEBUG( "receiver: bad CRC!\n" );
				if( (info->usHopInp) < MAX_HOP ) {
					//    hop- -  NAK    
					(info->usHopInp) ++;
					info->signal = SignalNAK;
					break;
				} else {
					//     hop-,     
					info->active_inbuf.signature |= S_DATA_NAK;
				}
			}

			info->usHopInp = 0;

			//  


			info->active_inbuf.adr = current_address;

			//    FIFO

#ifdef __DEBUG_TIMING__
			if(rs485_timeout(100L) ) {
				RS_PRINT("Long packet: > 100 usec!\n");
			}
#endif
			rs485_put_data(info);

			//    ACK
			info->signal = SignalACK;
			info->lSection = SECT_SIGNAL;

			break;

		case SECT_SIGNAL:

			rs485_read_enable(info);
			outb(OTHER, info->base +UART_LCR);	//    "  "
			SharedMaskInterrupt = UART_LSR_DR;

			if( ( iir & UART_LSR_PE ) !=0 ) {
				PDEBUG( "receiver: non signal in signal-sectiion!\n" );
			}

			//    ,     ...
			goto QUIT_HANDLER;

			break;

		}
		if( info->signal ) {
			//  
			outb(FIRST, info->base +UART_LCR);	//   "  "
			rs485_transmit_enable(info);
			outb(info->signal, info->base);
			info->lSection = SECT_SIGNAL;
			SharedMaskInterrupt = UART_LSR_TEMT;
		}
	} //   

QUIT_HANDLER:

//	down_flags(FRECIVER,&info->lFlags);

//#ifdef __DEBUG_TIMING__
//	rs485_gettime(&hand_fin);
//	rs485_printtm_interval( &hand_start, &hand_fin );
//#endif

#ifdef __DEBUG_SIGNAL__
	outb(RXENABLE, info->base +RX_TX_REG + 8);
#endif

//	rs485_allow_interrupt_card(info);

	return;
}
