#include <linux/errno.h>

#include <linux/sched.h>
#include <asm/uaccess.h>	// put_user

#include "mcua.h"
#include "rs485.h"
#include "rserrno.h"
#include "rs485_fifo.h"

void rs485_fifo_flush(RS485INFO *info) {
	info->rs->inbuffer_pas.adr = 0;
	info->rs->inbuffer_act.adr = 0;
	info->rs->fifo.data_wp = 0; info->rs->fifo.data_rp = 0;
	atomic_set( &(info->rs->fifo.data_filled), 0 );
}

void rs485_fifo_init(RS485INFO *info) {
	init_waitqueue_head( &(info->rs->fifo.rs485_queue) );
	rs485_fifo_flush(info);
}

/*
int rs485_fifo_create(RS485INFO *info) {
	return SUCCESS;
}

void rs485_fifo_destroy(void) {
	return;
}
*/
int inline rs_min( int val1, int val2 ) {
	if( val1 <= val2 ) return val1;
	else return val2;
}

int rs485_data_fifo_test( size_t count, RS485INFO* info ) {
	RS485FIFO* fifo = &(info->rs->fifo);

	if( (SIZE_STREAM_FIFO - atomic_read( &(fifo->data_filled) )) < count ) {
		RS_PRINT( KERN_CRIT "rs485_data_fifo_test: FIFO overflow! filled=%d count=%zd\n",
					atomic_read( &(fifo->data_filled) ), count );
		rs485_fifo_flush(info);
		return -ERSFIFOVERFLOW;
	} else {
		return SUCCESS;
	}
}

ssize_t rs485_data_fifo_put( void* buf0, size_t count, RS485INFO* info ) {
/*
 * Handler's function! (race condition with fifo_read)
 */
	RS485FIFO* fifo = &(info->rs->fifo);
	unsigned char* buf = (unsigned char*)buf0;
	unsigned int j;
	size_t count_writ=0;
	size_t count0 = count;
	int attempt = 0;
        int ret = SUCCESS;

        ret = rs485_data_fifo_test(count, info);
        if(ret < 0) return ret;

REPEAT:
//	left = (fifo->data_rp + SIZE_STREAM_FIFO -
//					fifo->data_wp ) % SIZE_STREAM_FIFO;
//	PDEBUG( "left is %i\n", left );
//	if( left == 1 ) {
//		RS_PRINT( KERN_CRIT "rs485_data_fifo_put: FIFO overflow! wp=%d rp=%d\n",
//					fifo->data_wp, fifo->data_rp );
//		return -ENOSPC;
//	}

	if( fifo->data_wp >= fifo->data_rp ) {
		count = rs_min( count, SIZE_STREAM_FIFO - fifo->data_wp );
//		if( count == left ) count--;
	} else {
// the write pointer has wrapped:
		count = rs_min( count,info->rs->fifo.data_rp - info->rs->fifo.data_wp );
	}
	for( j=0; j<count ; j++ ) {
		fifo->data_fifo[fifo->data_wp+j] = buf[j];
	}
	fifo->data_wp += count;
	count_writ += count;
	atomic_add( count, &(fifo->data_filled) );

	if( fifo->data_wp == SIZE_STREAM_FIFO ) {
		fifo->data_wp = 0;
	}

	if( count_writ < count0 ) {
		if( attempt > 0 ) {
			RS_PRINT( KERN_CRIT "rs485_data_fifo_put: FIFO overflow 2! attempt = %d\n",
						attempt );
			return -ENOSPC;
		}
		attempt++;
		buf = (unsigned char*)buf0+count;
		count = count0 - count;
//		PDEBUG( "put_data_fifo: repeat\n" );
		goto REPEAT;
	} else {
		return count_writ;
	}
}

ssize_t rs485_data_fifo_read( RS485INFO* info, unsigned char __user *buf, size_t count ) {
/*
 * NONBLOCKED read
 * copy to user space
 */
	RS485FIFO* fifo = &(info->rs->fifo);
	size_t count_read = 0;
	int attempt = 0;
	size_t count0 = count;
        unsigned long  rest = 0;

	if( atomic_read( &(fifo->data_filled) ) < count ) {
		return 0;
	}

REPEAT:
//	if( fifo->data_rp  == fifo->data_wp ) {
//		return 0;
//	}

	if( fifo->data_wp > fifo->data_rp ) {
		count = rs_min( count, fifo->data_wp - fifo->data_rp );
	} else {
// the write pointer has wrapped
		count = rs_min( count, SIZE_STREAM_FIFO - fifo->data_rp );
	}
	rest = copy_to_user( buf, fifo->data_fifo+fifo->data_rp, count);
	if(rest != 0) {
	    PDEBUG("rs485_data_fifo_read: copy_to_user rest = %ld\n",rest);
	    return -EFAULT;	
	}

	fifo->data_rp += count;
	count_read += count;
	atomic_sub( count, &(fifo->data_filled) );

	if( fifo->data_rp == SIZE_STREAM_FIFO ) {
		fifo->data_rp = 0;
	}

	if( count_read < count0 ) {
		if( attempt > 0 ) {
			RS_PRINT( "rs485_data_fifo_read: ERROR! attempt=%d\n",attempt );
			return 0;
		}
		attempt++;
		buf = buf+count;
		count = count0 - count;
//		PDEBUG( "read_data_fifo: repeat\n" );
		goto REPEAT;
	} else {
		return count_read;
	}
}

int rs485_put_data(RS485INFO *info) {

//  active   FIFO
// Proiskhodit v obrabotchike preryvaniya => nikto vklinit'sia ne smozhet

	unsigned int idev;
	unsigned char c_dev;
	int num;
	int err;

	if(info->rs->signal) {
/*
V fifo_put testiruetsia!
		if( rs485_data_fifo_test(sizeof(unsigned char), rs485) < 0 ) {
			PDEBUG( "rs485_put_data_fifo: FIFO overflow\n" );
			rs485_fifo_flush(rs485);
			return -ERSFIFOVERFLOW;
		}
*/
		err = rs485_data_fifo_put(&info->rs->signal,
					sizeof(unsigned char), info);
		info->rs->signal=0;
		if( err < 0 ) {
			PDEBUG( "rs485_put_data_fifo: err-code=%d\n",err );
                        return err;
		}
		else return SUCCESS; // err = nbytes
	}

	idev = ((unsigned int) info->rs->inbuffer_act.adr & MASK_ADR_PACKET);	// get address
        if(idev >= MAX_TASK_DEVICE) {    // TEST
            PDEBUG("rs485_put_data: idev >= MAX_TASK_DEVICE! addr = %x\n", info->rs->inbuffer_act.adr);
            return -ERSFORMAT;
        }
	num = info->rs->inbuffer_act.adr & MASK_CNT_PACKET;	// get packet number
	num /= 0x20; 	

	if( info->rs->device[idev].round_count_prev == num ) {	// this is a repeated packet?
	    info->rs->inbuffer_act.signature = 0;
	    return -ERSFORMAT;
	}

	if( num < info->rs->device[idev].round_count_prev ) {
		(info->rs->device[idev].number) += 4;
	}

	info->rs->device[idev].number &= ~3;
	info->rs->device[idev].number |= num;
	info->rs->device[idev].round_count_prev = num;

#ifndef __LPT__
	info->rs->inbuffer_act.size ++;	// real size
#else
// without CRC
	info->rs->inbuffer_act.size += 2;	// real size
#endif
	/* ----------------------------------------
         * format of the packet
         * +0 size
         * +1 signature
         * +2 device-address
         * +3 packet number (2 bytes)
         * +5 data
	 ------------------------------------------ */

	if( rs485_data_fifo_test(info->rs->inbuffer_act.size+1, info) < 0 ) {
		PDEBUG( "rs485_put_data_fifo: FIFO overflow\n" );
		rs485_fifo_flush(info);
		return -ERSFIFOVERFLOW;
	}

	err = rs485_data_fifo_put( &info->rs->inbuffer_act.size,
					sizeof(unsigned char), info );
	if( err < 0 ) {
		PDEBUG( "rs485_put_data_fifo: err-code=%d\n",err );
		return err;
	}

	err = rs485_data_fifo_put( &info->rs->inbuffer_act.signature,
					sizeof(unsigned char), info );

	if( err < 0 ) {
		PDEBUG( "rs485_put_data_fifo: err-code=%d\n",err );
		return err;
	}

	info->rs->inbuffer_act.signature = 0;
	c_dev = (unsigned char)idev;
	err = rs485_data_fifo_put( &c_dev,
					sizeof(unsigned char), info );
	if( err < 0 ) {
		PDEBUG( "rs485_put_data_fifo: err-code=%d\n",err );
		return err;
	}
	err = rs485_data_fifo_put( &info->rs->device[idev].number,
					sizeof(unsigned short), info );
	if( err < 0 ) {
		PDEBUG( "rs485_put_data_fifo: err-code=%d\n",err );
		return err;
	}
	err = rs485_data_fifo_put( &(info->rs->inbuffer_act.cache[2]),
							info->rs->inbuffer_act.size-4, info );

	if( err <= 0 ) {
		PDEBUG( "rs485_put_data_fifo: err-code=%d\n",err );
		return err;
	}

	return(SUCCESS);
}

