#include <mcua.h>
#ifndef __PP_PC__
#include <rs485.h>
#include <linux/interrupt.h>
#include <asm/irq.h>
#ifdef __CP132__
	#include <moxa132.h>
#endif
#ifdef __CP102__
	#include <moxa102.h>
#endif
#ifdef __COM__
	#include <com.h>
#endif
#ifdef __LPT__
	#include <lpt.h>
#endif
#include "rs485_irq.h"

extern RS485INFO *info_port1, *info_port2;

#ifndef KERN_2_6
static void rs485_interrupt(int irq, void *private, struct pt_regs *regs) {
#else
#ifdef KERN_2_6_21
static irqreturn_t rs485_interrupt(int irq, void *private) {
#else
static irqreturn_t rs485_interrupt(int irq, void *private, struct pt_regs *regs) {
#endif
#endif
	RS485INFO *info = (RS485INFO *) private;
        if(info->rs == NULL) {
            PDEBUG("rs485irq: empty interrupt %d!\n", irq);
#ifndef KERN_2_6
            return;
#else
            return IRQ_NONE;
#endif
        }
	receiv_data(info);
#ifndef KERN_2_6
	return;
#else
	return IRQ_HANDLED;
#endif
}

int rs485_request_irq(RS485INFO* info) {
	int err=0;
#ifndef KERN_2_6
	unsigned long flags;
	save_flags(flags); cli();
#endif
// SA_SHIRQ - PCI trebuet shared irq (2.6 -tochno, 2.4 - ne znayu)
	if( (err=request_irq(info->irq, rs485_interrupt,
#ifndef KERN_2_6_18
                                SA_INTERRUPT
#else
				IRQF_DISABLED
#endif
				, ADAPTER_ID,
				(void*)info)) != 0 ) {
		RS_PRINT(KERN_CRIT "rs485: request irq %d failed\n", info->irq);
	}
#ifndef KERN_2_6
	sti(); restore_flags(flags);
#endif

    return err;

}

void rs485_clean_irq(RS485INFO* info) {
#ifndef KERN_2_6	
	unsigned long flags;
	save_flags(flags); cli();
	free_irq(info->irq, info);
	sti(); restore_flags(flags);
#else
	free_irq(info->irq, info);
#endif
}


static unsigned long flags;
static int is_flags_saved = 0;

#ifndef KERN_2_6
void rs485_stop_interrupts(void) {
// disable all interrupts
	save_flags(flags);
	is_flags_saved = 1;
	cli();
}

void rs485_allow_interrupts(void) {
// enable interrupts
	sti();
	if(is_flags_saved) restore_flags(flags);
}
#else
void rs485_stop_interrupts(void) {
	is_flags_saved = 1;
	local_irq_save(flags);
}

void rs485_allow_interrupts(void) {
	if(is_flags_saved) local_irq_restore(flags);
}
#endif

#endif // !__PP_PC__
