#include <linux/module.h>

#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/time.h>

#include <mcua.h>
#include "rs485.h"
#include "rs485_tools.h"

static struct timeval start_waiting;

/*
int iround(float a) {
	int i;
	if(a >= INT_MAX) return(INT_MAX);
	if(a <= INT_MIN) return(INT_MIN);
	i = (int) a;
	if(a < (float)i + 0.5) return i;
	return(i+1); 
}
*/

inline void rs485_printtime(rs485_time_t *tm) {
	RS_PRINT( "rs485-time> %ld s %ld usec\n", tm->tv_sec, tm->tv_usec );
	return;
}

inline void rs485_printtm_interval(rs485_time_t *tm_start, rs485_time_t *tm_end) {
	RS_PRINT( "rs485-dtime> %ld usec\n",
                        (tm_end->tv_sec - tm_start->tv_sec)*1000000L+
                        (tm_end->tv_usec - tm_start->tv_usec) );
	return;
}

inline void rs485_set_starttime(void) {
	do_gettimeofday( &start_waiting );
}

inline void rs485_gettime(rs485_time_t *tm) {
	do_gettimeofday(tm);
	return;
}

inline int rs485_timeout(long interval) {	// usec
	struct timeval tv;
	long usec;

	do_gettimeofday(&tv);
	usec = (tv.tv_sec-start_waiting.tv_sec)*1000000L
					 + (tv.tv_usec-start_waiting.tv_usec);
	if( usec > interval ) {
            usec = (tv.tv_sec-start_waiting.tv_sec)*1000000L
                            + (tv.tv_usec-start_waiting.tv_usec);
            if( usec <= interval ) PDEBUG( "False timeout!\n" );
        }
        return ( usec > interval )? 1 : 0;
}

unsigned char crc(unsigned char *buf, size_t size) {
/*-------------------------------------------------
 *  calculation of the control sum (crc)
 *  return crc
 *  *buf - poiter to data
 *  size - buffer size
 --------------------------------------------------*/	
    unsigned short acc;
    unsigned short divisor = 0x700;
    unsigned int i,j;
    if( (size == 0) || (size > PACKET_MAX_LENGTH) ) return 0;
    *(buf+size) = 0x0;
    acc = (*(buf))<<8;
    for( j = 0; j < size; j++ ){ 
        acc = acc|(*(buf+j+1));
        for( i = 0; i < 8; i++ ){
            if( acc&0x8000 ) {
                acc = acc<<1;
                acc = acc^divisor;
            } else acc = acc<<1;
        }
    }
    acc = acc>>8;
    *(buf+size) = (unsigned char)acc;
    return( (unsigned char)acc );
}
